from pathlib import Path

import numpy as np
from mne.io.brainvision.brainvision import RawBrainVision
from mne.io.edf.edf import RawEDF, RawGDF
from mne.io.eeglab.eeglab import RawEEGLAB

from .Recording import Recording
from .Event import Event
from .RecordingMetadata import RecordingMetadata
from .constants import EEG_MONTAGES
from .errors import BreesyError, BreesyInternalError, protect_from_lib_error
from .load_events import find_and_load_event_for_file
from .type_hints import enforce_type_hints

# Note: if changing this, then update load_dataset if/else statement as well
SUPPORTED_FILE_EXTENSIONS = ['.gdf', '.bdf', '.eeg', '.csv', '.mat', '.tsv', '.vhdr', '.npy']


@enforce_type_hints
def load_dataset(filename: str, sample_rate: int | float | None = None) -> Recording:
    """Load a dataset from a file with a supported extension.

    :param filename: Path to the file to load
    :param sample_rate: Optional sampling rate in Hz

    :return: Recording object containing the loaded data
    """
    if not isinstance(filename, str):
        raise BreesyError("Invalid filename argument type",
                          "Please provide a string as the file name, example: 'my_data.gdf'")

    file_extension = Path(filename).suffix
    if not file_extension:
        raise BreesyError("File name should contain an extension",
                          "Provide a file with a supported extension, example: 'my_data.gdf'")

    is_supported_extension = file_extension in SUPPORTED_FILE_EXTENSIONS
    if not is_supported_extension:
        raise BreesyError(f"Unsupported file extension: {file_extension}",
                          "Provide a file with a supported extension, example: 'my_data.gdf'. "
                          f"Currently supported extensions: {SUPPORTED_FILE_EXTENSIONS}")

    path_filename = Path(filename)
    is_valid_filename = Path.is_file(path_filename)
    if not is_valid_filename:
        raise BreesyError(
            f"File not found: {filename}", "a) Make sure that the file is in your working directory; "
                                           f"the current working directory is: {Path.cwd()}. "
                                           f"b) Double check that your file explorer is showing the full file name "
                                           f"- Windows does not show file extensions by default.")

    if file_extension == '.gdf':
        return load_gdf(filename)
    elif file_extension == '.bdf':
        return load_bdf(filename)
    elif file_extension == '.eeg':
        # TODO: Warning, this might also be another .EEG supporting application! Maybe use more specific function
        #  EEGLab uses .set and .fdt files;
        #  Maybe can just fall back to MNE to decide what might be the authoring application?
        return load_eeglab_eeg(filename)
    elif file_extension == '.vhdr':
        return load_eeglab_eeg(filename)
    elif file_extension == '.csv':
        return load_csv(filename)
    elif file_extension == '.mat':
        return load_mat(filename, sample_rate=sample_rate)
    elif file_extension == '.tsv':
        return load_tsv(filename)
    elif file_extension == '.npy':
        return load_numpy(filename)
    else:
        raise BreesyInternalError("File extension valid but not handled")


@enforce_type_hints
def load_gdf(filename: str):
    """Load a GDF file.

    :param filename: Path to the GDF file

    :return: Recording object containing the loaded data
    """
    raw = _read_raw_gdf(filename)
    return _convert_mne_to_recording(raw, filename)

@enforce_type_hints
def load_bdf(filename: str):
    """Load a BDF file.

    :param filename: Path to the BDF file

    :return: Recording object containing the loaded data
    """
    raw = _read_raw_bdf(filename)
    return _convert_mne_to_recording(raw, filename)

@enforce_type_hints
def load_eeglab_eeg(filename: str):
    """Load an EEGLAB EEG file.

    :param filename: Path to the EEG file

    :return: Recording object containing the loaded data
    """
    raw = _read_raw_eeglab(filename)
    return _convert_mne_to_recording(raw, filename)


def _mat_extract_meaningful_keys(mat_data: dict) -> list[str]:
    keys = [k for k in mat_data.keys() if not k.startswith('__')]
    if not keys:
        raise BreesyError(
            "No valid data found. The MATLAB file appears to be empty or contains only metadata."
        )    
    elif len(keys) == 1:
        print(
            "Warning: data file likely contains only the recording and "
            "no useful metadata. Please find the sample rate and channel names "
            "in your data description."
        )  # TODO: fast instructions or link to documentation
    return keys


@enforce_type_hints
def load_mat(filename: str, sample_rate: int | float | None = None) -> Recording:
    """Load a MATLAB .mat file.

    :param filename: Path to the .mat file
    :param sample_rate: Sampling rate in Hz

    :return: Recording object containing the loaded data
    """
    mat_data = _loadmat(filename)

    meaningful_keys = _mat_extract_meaningful_keys(mat_data)

    data_key = None

    if len(meaningful_keys) == 1:
        data_key = meaningful_keys[0]
    else:
        # Find the largest array that looks like EEG data
        largest_size = 0
        for key in meaningful_keys:
            value = mat_data[key]
            if isinstance(value, np.ndarray) and value.size > largest_size and value.ndim >= 2:
                largest_size = value.size
                data_key = key

    data = mat_data[data_key]  # TODO: there can be 'O' types, as AI previously suggested?
    if not isinstance(data, np.ndarray):
        raise BreesyError(
            f"Problem reading data which is of type {type(data)}. It may be that Breesy failed to read the file correctly."  # TODO: test different files
        )
    if data.ndim != 2:
        raise BreesyError(
            "Found data has an incorrect number of dimensions. "
            "Expected data.ndim to be 2, but got "
            f"{data.ndim} dimensions instead.",
            "The data may be not EEG data, "
            "there are several records joined into one file, "
            "or Breesy did not found the correct data array in the file."
        )

    if data.shape[0] > data.shape[1]:
        print("Warning: the first data dimension seems to be the samples, because data contains more rows than columns. Dimensions will be switched places.")
        data = data.T
    n_channels, n_samples = data.shape
    default_channel_names = [f'Channel_{i + 1}' for i in range(data.shape[0])]

    # find or create channel names
    channel_name_keys = _find_keys_in_data(meaningful_keys, ['ch', 'ch name', 'ch names', 'channel names', 'channels', 'names', '10-20', '10-10', 'pos', 'positions'])
    if len(channel_name_keys) == 0:
        channel_names = default_channel_names
        channel_name_key = None
    elif len(channel_name_keys) > 1:
        print("Warning: "
            "found several keys in data which may contain channel names: "
            f"{channel_name_keys}. "
            "Using the first one.")        
        channel_name_key = channel_name_keys[0]
    else:
        channel_name_key = channel_name_keys[0]

    if channel_name_key:
        found_names = mat_data[channel_name_key]
        if not isinstance(found_names, (list, tuple, np.ndarray)):
            print("Warning: "
                "found channel name information does not appear to be a list: "
                f"{found_names}. "
                "Default channel names will be used instead.")
            channel_names = default_channel_names
        elif (n_channels != len(found_names)) and (n_samples != len(found_names)):
            print("Warning: "
                f"the number of channel names ({len(found_names)}) provided does not match the "
                f"number of channels ({n_channels}) in the EEG data. "
                "Default channel names will be used instead."
            )
            channel_names = default_channel_names
        elif (n_channels != len(channel_names)) and (n_samples == len(channel_names)):
            print('Warning: data rows and columns will be switched places according to the number of found channel names.')
            data = data.T
            n_channels, n_samples = data.shape
            channel_names = found_names
        else:
            channel_names = found_names

    if sample_rate:
        print(f"Overwriting sample rate to {sample_rate}")
        sample_rate = float(sample_rate)
    else:
        # try to find it in keys
        sample_rate = _detect_sample_rate_in_mat_keys(mat_data, meaningful_keys)
        if not sample_rate:
            # try to calculate it from time channel, if present
            sample_rate = _detect_sample_rate_from_time_column(data)
            if not sample_rate:
                raise BreesyError(
                    "No info found about sample rate. "
                    "Please find it in data description and set manually by "
                    "Pass an additional parameter 'sample_rate' into the data loading function."
                )

    data, channel_names = _leave_only_eeg_channels(data, channel_names)
    # channel names may not start from 1 after this, but maybe we should leave it like this

    metadata = RecordingMetadata(
        name=Path(filename).stem,
        file_path=Path(filename),
        description=f"MATLAB data loaded from {filename}"
    )

    return Recording(
        data=data,
        channel_names=channel_names,
        sample_rate=sample_rate,
        events=None,  # No events for now - could be added as a separate function
        metadata=metadata
    )


@enforce_type_hints
def load_delimited_text(filename: str, delimiter: str = ',', sample_rate: int | float | None = None) -> Recording:
    """Load a delimited text file (CSV, TSV, etc.).

    :param filename: Path to the text file
    :param delimiter: Character used to separate values in the file
    :param sample_rate: Sampling rate in Hz. If not provided, will attempt to guess from time column. If sample rate cannot be guessed, an error will be raised.

    :return: Recording object containing the loaded data
    """
    import csv

    # First pass: read headers and determine structure
    with open(filename, 'r', newline='') as file:
        reader = csv.reader(file, delimiter=delimiter)
        headers = next(reader)  # Get column names

        # Check if first column might be time
        time_column_idx = None
        if headers[0].lower() in ['time', 't', 'timestamp']:
            time_column_idx = 0

        # Exclude time column from channel names if present
        if time_column_idx is not None:
            channel_names = headers[1:]
        else:
            channel_names = headers

        # Count rows for pre-allocation
        row_count = sum(1 for _ in reader)

    # Second pass: read all data
    with open(filename, 'r', newline='') as file:
        reader = csv.reader(file, delimiter=delimiter)
        next(reader)  # Skip header

        # Pre-allocate data array
        data = np.zeros((len(channel_names), row_count))
        time_values = None

        if time_column_idx is not None:
            time_values = np.zeros(row_count)

        # Fill data
        for i, row in enumerate(reader):
            if i >= row_count:
                break

            # Convert values to float, handling potential errors
            try:
                row_values = [float(val) for val in row]
            except ValueError:
                # Skip rows that can't be converted to float
                continue

            if time_column_idx is not None:
                time_values[i] = row_values[time_column_idx]
                data_values = row_values[:time_column_idx] + row_values[time_column_idx + 1:]
            else:
                data_values = row_values

            for j, val in enumerate(data_values):
                if j < len(channel_names):
                    data[j, i] = val

    if sample_rate is not None:
        sample_rate = float(sample_rate)
    else:
        if time_values is not None and len(time_values) > 1:
            # Estimate sample rate
            time_diffs = np.diff(time_values)
            avg_diff = float(np.mean(time_diffs))
            if avg_diff > 0:
                sample_rate = float(round(1.0 / avg_diff))
        
        if sample_rate is None:
            raise BreesyError(
                "Sample rate was not provided and it could not be guessed from the data. "
                "Please provide the sample rate manually by passing the 'sample_rate' parameter.",
                "Provide the sample rate as a named argument to the function, e.g.: load_delimited_text('file.csv', sample_rate=500)"
            )

    # Create metadata
    metadata = RecordingMetadata(
        name=Path(filename).stem,
        file_path=Path(filename),
        description=f"Data loaded from {filename}"
    )

    # Look for event columns (columns with mostly 0s and few 1s)
    events = []
    for i, col_name in enumerate(channel_names):
        col_data = data[i, :]
        unique_vals = set(np.unique(col_data))
        # Check if column is binary and sparse (potential event marker)
        if unique_vals <= {0, 1} and np.mean(col_data) < 0.1:
            event_indices = np.where(col_data == 1)[0]
            for idx in event_indices:
                events.append(Event(name=col_name, index=idx))

    return Recording(
        data=data,
        channel_names=channel_names,
        sample_rate=sample_rate,
        events=events,
        metadata=metadata
    )

# Wrapper functions for backward compatibility
@enforce_type_hints
def load_csv(filename: str) -> Recording:
    """Load a CSV file.

    :param filename: Path to the CSV file

    :return: Recording object containing the loaded data
    """
    return load_delimited_text(filename, delimiter=',')


@enforce_type_hints
def load_tsv(filename: str) -> Recording:
    """Load a TSV file.

    :param filename: Path to the TSV file

    :return: Recording object containing the loaded data
    """
    return load_delimited_text(filename, delimiter='\t')


# Note: mostly LLM generated, looks reasonable
@enforce_type_hints
def load_numpy(filename: str, sample_rate: int | float | None = None) -> Recording:
    """Load a numpy .npy file.

    :param filename: Path to the .npy file
    :param sample_rate: Sampling rate in Hz. If None, defaults to 1000 Hz.

    :return: Recording object containing the loaded data
    """
    # Load the NumPy array with error protection
    data = _np_load(filename)

    # TODO: #12 we need to handle this case properly, currently just taking first trial
    if data.ndim == 3:
        # likely many recordings instead of one
        data = data[0]

    # Ensure it's a 2D array
    if data.ndim != 2:
        raise BreesyError(
            f"The NumPy array has {data.ndim} dimensions, but 2 dimensions are required.",
            "Make sure the NumPy file contains a 2D array with shape (channels, samples)."
        )

    # If the first dimension is larger, assume it's (samples, channels) and transpose
    if data.shape[0] > data.shape[1]:
        print("Warning: the first data dimension seems to be the samples, because data contains more rows than columns. Dimensions will be switched places.")
        data = data.T

    # Create default channel names
    channel_names = [f'Channel_{i+1}' for i in range(data.shape[0])]

    if sample_rate is None:
        print("Warning: No sample rate provided. Using default of 1000 Hz.")
        sample_rate = 1000
    else:
        sample_rate = float(sample_rate)

    # Create metadata
    metadata = RecordingMetadata(
        name=Path(filename).stem,
        file_path=Path(filename),
        description=f"NumPy data loaded from {filename}"
    )

    # Find and load events if available
    events = find_and_load_event_for_file(filename, sample_rate)

    return Recording(
        data=data,
        channel_names=channel_names,
        sample_rate=sample_rate,
        events=events,
        metadata=metadata
    )

# Note: LLM generated, looks reasonable
@enforce_type_hints
def load_brainvision_eeg(filename: str) -> Recording:
    """Load a BrainVision EEG file.

    :param filename: Path to the BrainVision file (.vhdr)

    :return: Recording object containing the loaded data
    """
    file_path = Path(filename)

    # Ensure the file exists
    if not file_path.exists():
        raise BreesyError(
            f"File not found: {filename}", "a) Make sure that the file is in your working directory; "
                                           f"the current working directory is: {Path.cwd()}. "
                                           f"b) Double check that your file explorer is showing the full file name "
                                           f"- Windows does not show file extensions by default.")

    # Determine which file we're dealing with
    if file_path.suffix.lower() == '.eeg':
        # If .eeg file provided, look for the corresponding .vhdr
        vhdr_path = file_path.with_suffix('.vhdr')
        if not vhdr_path.exists():
            raise BreesyError(
                f"Header file not found for {filename}",
                "BrainVision files require a .vhdr file. Make sure it's in the same directory as the .eeg file."
            )
        filename = str(vhdr_path)
    elif file_path.suffix.lower() != '.vhdr':
        raise BreesyError(
            f"Unsupported BrainVision file extension: {file_path.suffix}",
            "Please provide a .vhdr or .eeg file."
        )

    # Load the EEG data using MNE
    raw = _read_raw_brainvision(filename)

    return _convert_mne_to_recording(raw, filename)

@enforce_type_hints
def load_all_files_in_folder_by_extension(file_extension: str, folder_path: str = ""):
    """Load all files with a specific extension from a folder.

    :param file_extension: File extension to filter by (e.g., '.mat')
    :param folder_path: Path to the folder containing the files

    :return: List of Recording objects containing the loaded data
    """
    if not isinstance(folder_path, str):
        raise BreesyError("Invalid folder_path argument type",
                          "Provide a string as the folder path, example: 'data_folder', or '' (empty string or period) for current working directory.")

    if not isinstance(file_extension, str):
        raise BreesyError("Invalid file_extension argument type",
                          "Provide a string as the file extension, example: '.gdf'")

    folder = Path(folder_path)
    if not folder.is_dir():
        raise BreesyError(f"The folder path was not found: {folder.resolve()}",
                          f"Check the folder path manually to make sure that the folder actually exists.")

    full_path = folder.resolve()

    files = list(full_path.glob(f"*{file_extension}"))
    if not files:
        raise BreesyError(
            f"No {file_extension} files found in folder: '{folder_path}' (resolved absolute path: {folder.resolve()})",
            f"Check the folder to make sure that it actually contains {file_extension} files.")

    data_list = []
    for file in files:
        data = load_dataset(str(file))
        data_list.append(data)

    return data_list


@enforce_type_hints
def load_all_mat_files_in_folder(folder_path: str = ""):
    """Load all MATLAB .mat files from a folder.

    :param folder_path: Path to the folder containing the files

    :return: List of Recording objects containing the loaded data
    """
    return load_all_files_in_folder_by_extension(".mat", folder_path)


@enforce_type_hints
def load_all_gdf_files_in_folder(folder_path: str = ""):
    """Load all GDF files from a folder.

    :param folder_path: Path to the folder containing the files

    :return: List of Recording objects containing the loaded data
    """
    return load_all_files_in_folder_by_extension(".gdf", folder_path)


@enforce_type_hints
def load_all_bdf_files_in_folder(folder_path: str = ""):
    """Load all BDF files from a folder.

    :param folder_path: Path to the folder containing the files

    :return: List of Recording objects containing the loaded data
    """
    return load_all_files_in_folder_by_extension(".bdf", folder_path)


@enforce_type_hints
def load_all_eeg_files_in_folder(folder_path: str = ""):
    """Load all EEG files from a folder.

    :param folder_path: Path to the folder containing the files

    :return: List of Recording objects containing the loaded data
    """
    return load_all_files_in_folder_by_extension(".eeg", folder_path)


@enforce_type_hints
def load_all_csv_files_in_folder(folder_path: str = ""):
    """Load all CSV files from a folder.

    :param folder_path: Path to the folder containing the files

    :return: List of Recording objects containing the loaded data
    """
    return load_all_files_in_folder_by_extension(".csv", folder_path)


@enforce_type_hints
def load_all_tsv_files_in_folder(folder_path: str = ""):
    """Load all TSV files from a folder.

    :param folder_path: Path to the folder containing the files

    :return: List of Recording objects containing the loaded data
    """
    return load_all_files_in_folder_by_extension(".tsv", folder_path)

@enforce_type_hints
def load_all_brainvision_files_in_folder(folder_path: str = ""):
    """Load all BrainVision files from a folder.

    :param folder_path: Path to the folder containing the files

    :return: List of Recording objects containing the loaded data
    """
    return load_all_files_in_folder_by_extension(".eeg", folder_path)


@enforce_type_hints
def load_all_numpy_files_in_folder(folder_path: str = ""):
    """Load all NumPy .npy files from a folder.

    :param folder_path: Path to the folder containing the files

    :return: List of Recording objects containing the loaded data
    """
    return load_all_files_in_folder_by_extension(".npy", folder_path)

# --------- Helper functions ---------

@protect_from_lib_error("numpy")
def _np_load(filename):
    return np.load(filename)

@protect_from_lib_error("mne")
def _read_raw_gdf(filename):
    from mne.io import read_raw_gdf
    return read_raw_gdf(filename, verbose=False)

@protect_from_lib_error("mne")
def _read_raw_bdf(filename):
    from mne.io import read_raw_bdf
    return read_raw_bdf(filename, verbose=False)

@protect_from_lib_error("mne")
def _read_raw_eeglab(filename):
    from mne.io import read_raw_eeglab
    return read_raw_eeglab(filename, verbose=False)

@protect_from_lib_error("mne")
def _read_raw_brainvision(filename: str):
    from mne.io import read_raw_brainvision
    return read_raw_brainvision(filename, verbose=False)

@protect_from_lib_error("scipy")
def _loadmat(filename):
    from scipy.io import loadmat
    return loadmat(filename, squeeze_me=True)

def _convert_mne_to_recording(raw: RawGDF | RawEDF | RawEEGLAB | RawBrainVision, filename):
    data = raw.get_data()
    sample_rate = int(raw.info['sfreq'])
    channel_names = raw.ch_names

    # Extract events if available
    events = []
    try:
        # Use MNE to find events
        from mne import find_events
        mne_events = find_events(raw)
        for event in mne_events:
            sample_idx = event[0]
            event_id = event[2]
            event_name = raw.event_id.get(event_id, f"Event_{event_id}")
            events.append(Event(name=event_name, index=sample_idx))
    except Exception as e:
        pass
    if not events:
        print("Events not found through mne.find_events")

    # Try from annotations
    if not events:
        try:
            from mne import events_from_annotations
            mne_events, event_id = events_from_annotations(raw)
            # Map event IDs to event names
            id_to_name = {v: k for k, v in event_id.items()}

            for event in mne_events:
                sample_idx = event[0]
                event_code = event[2]
                event_name = id_to_name.get(event_code, f"Event_{event_code}")
                events.append(Event(name=event_name, index=sample_idx))
        except Exception as e:
            pass
    if not events:
        print("Events not found through mne.events_from_annotations")

    # Extract events for peer files
    # TODO: maybe this needs to be used in ALL load functions
    if not events:
        try:
            events = find_and_load_event_for_file(filename, sample_rate)
        except Exception as e:
            pass
    if not events:
        print("Events not found in related files")

    metadata = RecordingMetadata(
        name=Path(filename).stem,
        file_path=Path(filename),
        description=f"Data loaded from {filename}"
    )

    return Recording(
        data=data,
        channel_names=channel_names,
        sample_rate=sample_rate,
        events=events,
        metadata=metadata
    )


def _detect_sample_rate_in_mat_keys(mat_data: dict, keys: list[str]) -> float | None:
    sample_rate_keys = _find_keys_in_data(keys, ['sample rate', 'sp', 'sr', 'hz', 'sampling rate', 'sfreq', 'fs', 'srate'])
    if len(sample_rate_keys) == 0:
        return None

    sample_rate_key = sample_rate_keys[0]
    if len(sample_rate_keys) > 1:
        print("Warning: "
            "found several keys in data which may contain sample rate:"
            f"\n{sample_rate_keys}\n"
            "Using the first one.")

    sr_value = mat_data[sample_rate_key]
    try:
        if isinstance(sr_value, (int, float, str)):
            return float(sr_value)
        elif isinstance(sr_value, np.ndarray) and sr_value.size == 1:
            return float(sr_value.item())
        elif isinstance(sr_value, (list, tuple)) and len(sr_value) == 1:
            return float(sr_value[0])
        else:
            raise ValueError
    except ValueError:
        print("Warning: "
            f"information found in column {sample_rate_key} "
            "could not be read automatically. Mentioned column contents are:"
            f"\n{sr_value}\n"
            "Sample rate was not set."
        )
        return None


def _detect_sample_rate_from_time_column(data: np.ndarray) -> float | None:
    for i in range(data.shape[0]):
        if _is_likely_time_column(data[i]):
            time_step = float(data[i, 1] - data[i, 0])
            sample_rate = float(1 / time_step)
            return sample_rate
    return None


def _is_likely_eeg_column(column: np.ndarray) -> bool:
    return (not _column_is_monotonic(column)) and (not _column_is_sparse(column))


def _is_likely_time_column(column: np.ndarray) -> bool:
    u_diffs = np.unique(np.diff(column).round(6))
    if len(u_diffs) == 1 and u_diffs[0] > 0:
        return True
    return False


def _leave_only_eeg_channels(data: np.ndarray, ch_names: list[str]) -> tuple[np.ndarray, list[str]]:
    eeg_ix = [i for i in range(data.shape[0]) if _is_likely_eeg_column(data[i])]
    new_channel_names = [x for i, x in enumerate(ch_names) if i in eeg_ix]
    return data[eeg_ix], new_channel_names


def _column_has_eeg_name(column_name: str) -> bool:
    # TODO: #13 should not be case-sensitive?
    return column_name in EEG_MONTAGES['10-10']


def _column_is_monotonic(column: np.ndarray) -> bool:
    diffs = np.diff(column)
    return bool(np.all(diffs >= 0))


def _column_is_sparse(column: np.ndarray, sparsity_threshold: float = 0.2) -> bool:
    u_values, u_counts = np.unique(column, return_counts=True)
    most_common_value_rate = np.max(u_counts) / len(column)
    return bool(most_common_value_rate > sparsity_threshold)


def _find_keys_in_data(keys: list[str], searching_for: list[str]) -> list:
    found = []
    translation = str.maketrans('', '', '_- .')
    to_check = [k.lower().translate(translation) for k in searching_for]
    for k in keys:
        if k.lower().translate(translation) in to_check:
            found.append(k)
    return found



from pathlib import Path

from .type_hints import enforce_type_hints


class RecordingMetadata:
    """Metadata for an EEG recording."""

    name: str
    """Name of the recording"""

    file_path: Path
    """Path to the source file"""

    description: str
    """Optional description of the recording"""

    def __str__(self):
        """Return a string representation of the RecordingMetadata object."""
        return self.__repr__()

    def __repr__(self):
        """Return a string representation of the RecordingMetadata object."""
        description_repr = f", description='{self.description}'" if self.description else ""
        return f"RecordingMetadata(name='{self.name}', file_path='{self.file_path}'{description_repr})"

    @enforce_type_hints
    def __init__(self, name: str, file_path: Path, description: str | None = None):
        """Initialize RecordingMetadata.

        :param name: Name of the recording
        :param file_path: Path to the source file
        :param description: Optional description of the recording
        """
        self.name = name
        self.file_path = file_path
        self.description = description



from sklearn.decomposition import FastICA

from breesy.errors import protect_from_lib_error


@protect_from_lib_error("sklearn")
def _sklearn_fastica(n_components: int, random_state: int | None = None) -> FastICA:
    return FastICA(n_components=n_components, whiten='unit-variance',
                   max_iter=1000, random_state=random_state)


# TODO: add picard ICA (from the picard package)


from __future__ import annotations  # for self references in return types
from collections import abc
from typing import TypeVar, Generic, Optional, Mapping, Iterable, Tuple, Union, Dict, overload, Iterator, Any

# Adapted from: https://github.com/DeveloperRSquared/case-insensitive-dict/blob/main/case_insensitive_dict/case_insensitive_dict.py

KT = TypeVar('KT')  # pylint: disable=invalid-name
VT = TypeVar('VT')  # pylint: disable=invalid-name
MutableMapping = abc.MutableMapping[KT, VT]  # pylint: disable=unsubscriptable-object

class CaseInsensitiveDict(MutableMapping, Generic[KT, VT]):
    @overload
    def __init__(self, data: Optional[Mapping[KT, VT]] = None) -> None:
        ...

    @overload
    def __init__(self, data: Optional[Iterable[Tuple[KT, VT]]] = None) -> None:
        ...

    def __init__(self, data: Optional[Union[Mapping[KT, VT], Iterable[Tuple[KT, VT]]]] = None) -> None:
        # Mapping from lowercased key to tuple of (actual key, value)
        self._data: Dict[KT, Tuple[KT, VT]] = {}
        if data is None:
            data = {}
        self.update(data)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({dict(self.items())!r})'

    @staticmethod
    def _convert_key(key: KT) -> KT:
        if isinstance(key, str):
            return key.lower()  # type: ignore[return-value]
        return key

    def _get_key_value(self, key: KT) -> Tuple[KT, VT]:
        try:
            return self._data[self._convert_key(key=key)]
        except KeyError:
            raise KeyError(f"Key: {key!r} not found.") from None

    def __setitem__(self, key: KT, value: VT) -> None:
        self._data[self._convert_key(key=key)] = (key, value)

    def __getitem__(self, key: KT) -> VT:
        return self._get_key_value(key=key)[1]

    def __delitem__(self, key: KT) -> None:
        del self._data[self._convert_key(key=key)]

    def __iter__(self) -> Iterator[KT]:
        return (key for key, _ in self._data.values())

    def __len__(self) -> int:
        return len(self._data)

    def lower_items(self) -> Iterator[Tuple[KT, VT]]:
        return ((key, val[1]) for key, val in self._data.items())

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, abc.Mapping):
            return False
        other_dict = CaseInsensitiveDict[Any, Any](data=other)
        return dict(self.lower_items()) == dict(other_dict.lower_items())

    def copy(self) -> CaseInsensitiveDict[KT, VT]:
        return CaseInsensitiveDict(data=dict(self._data.values()))

    def getkey(self, key: KT) -> KT:
        return self._get_key_value(key=key)[0]

    @classmethod
    def fromkeys(cls, iterable: Iterable[KT], value: VT) -> CaseInsensitiveDict[KT, VT]:
        return cls([(key, value) for key in iterable])



from functools import wraps

# The main difference from Exception is that we provide a suggested fix
class BreesyError(Exception):
    def __init__(self, message: str, suggested_fix: str | None = None):
        super().__init__(message)
        self.suggested_fix = suggested_fix

    def __str__(self):
        if self.suggested_fix:
            return f"\n ⚠️ {super().__str__()} \n 💡 SUGGESTED FIX: {self.suggested_fix}"
        return super().__str__()


# In case a library throws and we didn't anticipate it, we want to protect the user from a scary message
class BreesyLibraryError(BreesyError):
    def __init__(self, library_name: str, ex: Exception | None = None):
        super().__init__(
            f"Oh no, a library that Breesy is using threw an error! Please show this error to Breesy developers! "
            f"Error with library named '{library_name}', error message: '{ex}'",
            f"Google or ask an AI the library name and the error message, you will probably find some useful advice. "
            f"Try other ways of achieving your goal with Breesy, ask colleagues or the lecturer for help. "
            f"Example query to Google/AI: 'explain fix {library_name} exception {ex}'")


# Breesy error specifically for cases where the user is not at fault
class BreesyInternalError(BreesyError):
    def __init__(self, message: str, ex: Exception | None = None):
        super().__init__(
            "Oh no, something unexpected went wrong! Please show this error to Breesy developers (your lecturer)!"
            f"Error message: '{message}';"
            f"Internal error: '{ex}'",
            "Try other ways of achieving your goal with Breesy, ask colleagues or the lecturer for help.")


# Decorator to protect the user from library errors
def protect_from_lib_error(library_name: str):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as ex:
                raise BreesyLibraryError(library_name, ex)

        return wrapper

    return decorator



"""Breesy - Brainwave Retrieval, Extraction, and Easy Signal Yield.

Breesy provides easy-to-use functions and pipelines for neuroscience students
who want to apply signal processing and statistical methods on EEG data.
"""

__version__ = '0.1.0'

# Import core classes
from .Recording import Recording
from .Event import Event
from .RecordingMetadata import RecordingMetadata

# Make key modules and functions available at package level
from . import events, load, constants, features, type_hints, errors, \
    signal_generation, processing, plots, transform, data, load_events



import numpy as np

from .Event import Event
from .RecordingMetadata import RecordingMetadata
from .errors import BreesyError
from .type_hints import enforce_type_hints


class Recording:
    """Represents an EEG recording with data, channel information, and events."""

    data: np.ndarray
    """EEG data array of shape (channels, samples)"""

    channel_names: list[str]
    """List of channel names"""

    sample_rate: int | float
    """Sampling rate in Hz"""

    events: list[Event] | None
    """List of Event objects"""

    metadata: RecordingMetadata | None
    """Additional metadata about the recording"""

    @enforce_type_hints
    def __init__(self,
                 data: np.ndarray,
                 channel_names: list[str],
                 sample_rate: int | float,
                 events: list[Event] | None = None,
                 metadata: RecordingMetadata | None = None, ):
        """Initialize a Recording.

        :param data: EEG data array of shape (channels, samples)
        :param channel_names: List of channel names
        :param sample_rate: Sampling rate in Hz
        :param events: List of Event objects
        :param metadata: Additional metadata about the recording
        """

        if data.ndim != 2:
            raise BreesyError(
                "The EEG data provided has an incorrect number of dimensions. Expected data.ndim to be 2, but got "
                f"{data.ndim} dimensions instead.",
                "The data provided is likely not EEG data, or there are several records joined into one file. However, Breesy only supports ordinary two-dimensional EEG data for now."
            )

        n_channels, n_samples = data.shape
        if (n_channels != len(channel_names)) and (n_samples != len(channel_names)):
            raise BreesyError(
                f"The number of channel names ({len(channel_names)}) provided does not match the "
                f"number of channels ({n_channels}) in the EEG data.",
                "Make sure that the number of channel names provided matches the number of channels in the EEG data."
                "You can see the number of channel names like this: len(eeg_channel_names)."
            )
        elif (n_channels != len(channel_names)) and (n_samples == len(channel_names)):
            print('Warning: data rows and columns will be switched places according to the provided channel names.')
            self.data = data.T
        else:
            self.data = data

        self.channel_names = channel_names

        if sample_rate <= 0:
            raise BreesyError(
                f"Sample rate should be more than 0, but you provided {sample_rate}.",
                "Make sure that the sample rate is a positive integer. E.g., a sample rate of: 512."
            )

        is_typical_sample_rate = 100 < sample_rate < 5000
        if not is_typical_sample_rate:
            print(f"Warning: Unusual sample rate ({sample_rate} Hz). Typical EEG sample rates range from 128-2000 Hz.")
        self.sample_rate = sample_rate

        if events is None:
            events = []
        self._set_events(events)

        self.metadata = metadata

    def _set_events(self, events: list[Event]) -> None:
        out_of_range_events = [event for event in events if event.index >= self.number_of_samples]
        if out_of_range_events:
            raise BreesyError(
                f"Found {len(out_of_range_events)} events with indices out of range for the provided EEG data. "
                f"First Got index {out_of_range_events[0].index} while the EEG data has {self.number_of_samples} samples.",
                "Make sure that the event indices are within the range of the EEG data. "
                "For instance, if the EEG data has 10000 samples, the event indices should be less than 10000. "
                "You can see the number of samples in the EEG data like so: eeg_data.shape[1]."
            )

        sorted_events = sorted(events, key=lambda event: event.index)
        self.events = sorted_events

    def __repr__(self):
        """Return a string representation of the Recording object."""
        metadata_repr = f", name={self.metadata.name}" if self.metadata else ""
        events_repr = f", event_count={len(self.events)}" if self.events else ""
        return (f"Recording("
                f"channels={self.number_of_channels}, "
                f"duration={self.duration:.2f} s, "
                f"sample_rate={self.sample_rate} Hz"
                f"{metadata_repr}"
                f"{events_repr}"
                f")")


    @property
    def number_of_channels(self):
        """Return the number of channels in the recording."""
        return self.data.shape[0]

    @property
    def duration(self):
        """Return the duration of the recording in seconds."""
        return self.data.shape[1] / self.sample_rate

    @property
    def number_of_samples(self):
        """Return the number of samples in the recording."""
        return self.data.shape[1]



from breesy.load import load_mat

filename = "tests/matlab_dataset_karolina/FADHD.mat"
filename2 = "tests/matlab_dataset_karolina/MADHD.mat"

file = load_mat(filename)
print("====== Breesy file 1 ======")
print(file)

file2 = load_mat(filename2)
print("====== Breesy file 2 ======")
print(file2)


import numpy as np

from breesy.Event import Event
from breesy.Recording import Recording
from breesy.errors import BreesyError
from breesy.type_hints import enforce_type_hints
from breesy.breesy_scipy import _si_butter_bandpass_sos, _si_sosfiltfilt


@enforce_type_hints
def take_n_seconds(recording: Recording, duration: float | int = 5.0):
    """Take the first N seconds from a recording.

    :param recording: Input recording
    :param duration: Number of seconds to take
    
    :return: New recording with only the first N seconds
    """
    return cut_by_second_range(recording, 0, duration)


@enforce_type_hints
def cut_by_second_range(recording: Recording, start_second: float | int, end_second: float | int):
    """Cut a recording to a specific time range.

    :param recording: Input recording
    :param start_second: Start time in seconds
    :param end_second: End time in seconds
    
    :return: New recording containing only the specified time range
    """
    if start_second > end_second:
        raise BreesyError(
            f"Requested to cut from {start_second} s to {end_second} s, but the start time is higher than the end time.",
            "Make sure that the start time is lower than the end time."
        )

    if end_second > recording.duration:
        raise BreesyError(
            f"Requested to cut from {start_second} s to {end_second} s, but the recording has only {recording.duration}s.",
            "Make sure that the start and end seconds are within the range of the recording duration. "
            "You can see the recording duration like so: recording.duration."
        )

    start_index = recording.sample_rate * start_second
    end_index = recording.sample_rate * end_second

    return cut_by_sample_range(recording, start_index, end_index)


@enforce_type_hints
def take_n_samples(recording: Recording, n_samples: int) -> Recording:
    """Take the first N samples from a recording.

    :param recording: Input recording
    :param n_samples: Number of samples to take
    
    :return: New recording with only the first N samples
    """
    if n_samples > recording.number_of_samples:
        raise BreesyError(
            f"Requested to take {n_samples} samples, but the recording has only {recording.number_of_samples} samples.",
            "Make sure that the number of samples requested is less than the number of samples in the recording. "
            "You can see the number of samples in the recording like so: recording.num_samples."
        )

    return cut_by_sample_range(recording, 0, n_samples - 1)


@enforce_type_hints
def cut_by_sample_range(recording: Recording, start_index: int, end_index: int) -> Recording:
    """Cut a recording to a specific sample range.

    :param recording: Input recording
    :param start_index: Start sample index
    :param end_index: End sample index
    
    :return: New recording containing only the specified sample range
    """
    if start_index < 0 or end_index > recording.number_of_samples:
        raise BreesyError(
            f"Requested to cut from {start_index} to {end_index}, but the recording has only {recording.number_of_samples} samples.",
            "Make sure that the start and end sample indices are within the range of the recording samples. "
            "You can see the number of samples in the recording like so: recording.number_of_samples."
        )

    # Take both ends inclusive so that the events themselves also fit in the epoch
    new_eeg_data = recording.data[:, start_index:end_index + 1]
    events_in_new_data = [event for event in (recording.events or []) if start_index <= event.index <= end_index]
    events_with_offset = [Event(e.name, e.index - start_index) for e in events_in_new_data]

    return Recording(
        data=new_eeg_data,
        channel_names=recording.channel_names,
        sample_rate=recording.sample_rate,
        events=events_with_offset,
        metadata=recording.metadata
    )


@enforce_type_hints
def update_data(recording: Recording, new_eeg_data: np.ndarray) -> Recording:
    """Create a new recording with updated EEG data.

    :param recording: Input recording
    :param new_eeg_data: New EEG data array to use
    
    :return: New recording with updated data but same metadata and events
    """
    if recording.number_of_samples != new_eeg_data.shape[1]:
        raise BreesyError(
            f"New EEG data has a different number of samples ({new_eeg_data.shape[1]}) than the original EEG data ({recording.number_of_samples}).",
            "Make sure that the new EEG data has the same number of samples as the original EEG data. "
            "You can see the number of samples in the original EEG data like so: recording.number_of_samples."
        )

    return Recording(
        data=new_eeg_data,
        channel_names=recording.channel_names,
        sample_rate=recording.sample_rate,
        events=recording.events,
        metadata=recording.metadata
    )


@enforce_type_hints
def add_events(recording: Recording, events: list[Event]) -> Recording:
    """Create a new recording with additional events.

    :param recording: Input recording
    :param events: List of events to add
    
    :return: New recording with added events
    """
    new_events = recording.events or [] + events
    return Recording(
        data=recording.data,
        channel_names=recording.channel_names,
        sample_rate=recording.sample_rate,
        events=new_events,
        metadata=recording.metadata
    )


@enforce_type_hints
def remove_channels(recording: Recording, channels_to_remove: list[str]) -> Recording:
    """Create a new recording with specified channels removed.

    :param recording: Input recording
    :param channels_to_remove: List of channel names to remove
    
    :return: New recording without the specified channels
    """
    missing_channels = [channel for channel in channels_to_remove if channel not in recording.channel_names]
    if missing_channels:
        raise BreesyError(
            f"Channels to remove contain channel names that are not present in the recording, namely: {missing_channels}.",
            "Make sure that the channel names provided to remove are present in the recording. "
            "Double check that you haven't removed those channels for this data already. "
            "You can see the channel names in the recording with: recording.channel_names."
        )

    channel_indices_to_remove = [recording.channel_names.index(channel) for channel in channels_to_remove]
    new_eeg_data = np.delete(recording.data, channel_indices_to_remove, axis=0)
    new_channel_names = [channel for channel in recording.channel_names if channel not in channels_to_remove]
    return Recording(
        data=new_eeg_data,
        channel_names=new_channel_names,
        sample_rate=recording.sample_rate,
        events=recording.events,
        metadata=recording.metadata
    )


@enforce_type_hints
def select_channels(recording: Recording, channels_to_select: list[str]) -> Recording:
    """Create a new recording with only the specified channels.

    :param recording: Input recording
    :param channels_to_select: List of channel names to keep
    
    :return: New recording with only the specified channels
    """
    missing_channels = [channel for channel in channels_to_select if channel not in recording.channel_names]
    if missing_channels:
        raise BreesyError(
            f"Channels to select contain channel names that are not present in the recording, namely: {missing_channels}.",
            "Make sure that the channel names provided to select are present in the recording. "
            "You can see the channel names in the recording with: recording.channel_names."
        )

    channel_indices_to_select = [recording.channel_names.index(channel) for channel in channels_to_select]
    new_eeg_data = recording.data[channel_indices_to_select]
    return Recording(
        data=new_eeg_data,
        channel_names=channels_to_select,
        sample_rate=recording.sample_rate,
        events=recording.events,
        metadata=recording.metadata
    )


@enforce_type_hints
def set_channel_names(recording: Recording, ch_names: list[str]) -> Recording:
    """Set new channel names for a recording.

    :param recording: Input recording
    :param ch_names: List of new channel names
    
    :return: New recording with updated channel names
    """
    if len(ch_names) != recording.number_of_channels:
        raise BreesyError(
            f"{len(ch_names)} channel names are provided, but data has {recording.number_of_channels} channels. "
            "Please recheck your list of channel names. "
        )
    return Recording(
        data=recording.data,
        channel_names=ch_names,
        sample_rate=recording.sample_rate,
        events=recording.events,
        metadata=recording.metadata
    )


@enforce_type_hints
def get_events_by_name(recording: Recording, event_name: str) -> list[Event]:
    """Return a list of events with a specific name.

    :param recording: Input recording
    :param event_name: Name of events to find
    
    :return: List of events with the specified name
    """
    return [event for event in recording.events if event.name == event_name]


@enforce_type_hints
def split_by_continuous_events(recording: Recording, start_event_name: str, end_event_name: str) -> list[Recording]:
    """Split a recording into segments between start and end events.

    :param recording: Input recording
    :param start_event_name: Name of events marking segment starts
    :param end_event_name: Name of events marking segment ends
    
    :return: List of recording segments between each start-end event pair
    """
    if not recording.events:
        raise BreesyError(
            "No events found in the recording.",
            "Make sure that the recording has events. You can see the number of events like so: len(recording.events)."
        )

    start_events = get_events_by_name(recording, start_event_name)
    end_events = get_events_by_name(recording, end_event_name)

    if len(start_events) != len(end_events):
        raise BreesyError(
            f"Found {len(start_events)} start events and {len(end_events)} end events. "
            "Expected the number of start and end events to be the same.",
            "Make sure each start event has a corresponding end event. "
            "You can see the number of start events like so: get_events_by_name(recording, start_event_name); "
            "number of end events: get_events_by_name(recording, end_event_name)."
        )

    epoch_recordings = []
    for start_event, end_event in zip(start_events, end_events):
        if start_event.index >= end_event.index:
            pair_number = start_events.index(start_event) + 1
            raise BreesyError(
                f"When going pairwise through events, found a pair (nr. {pair_number}) where start event index ({start_event.index}) is higher than the end event index ({end_event.index}).",
                "Inspect the events manually, for start events: get_events_by_name(recording, start_event_name); "
                "end events: get_events_by_name(recording, end_event_name). "
                "Make sure that the start event index is lower than the end event index. "
            )

        epoch_recording = cut_by_sample_range(recording, start_event.index, end_event.index)
        epoch_recordings.append(epoch_recording)

    return epoch_recordings


@enforce_type_hints
def split_by_events(recording: Recording, event_name: str, duration: float | int = 5.0) -> list[Recording]:
    """Split a recording into segments starting at specified events.

    :param recording: Input recording
    :param event_name: Name of events to use as segment starts
    :param duration: Duration of each segment in seconds
    
    :return: List of recording segments starting at each event
    """
    events = get_events_by_name(recording, event_name)

    epoch_recordings = []
    for event in events:
        end_index = event.index + recording.sample_rate * duration
        epoch_recording = cut_by_sample_range(recording, event.index, end_index)
        epoch_recordings.append(epoch_recording)

    return epoch_recordings


@enforce_type_hints
def split_by_window_duration(recording: Recording, window_duration: float | int, overlap_duration: float | int = 0.0) -> list[Recording]:
    """Split a recording into equal-sized segments of specified duration with optional overlap.

    :param recording: Input recording
    :param window_duration: Duration of each segment in seconds
    :param overlap_duration: Duration of overlap between consecutive segments in seconds (default: 0.0)
    
    :return: List of equal-sized recording segments
    """
    if overlap_duration < 0:
        raise BreesyError(
            f"Overlap duration ({overlap_duration}s) cannot be negative.",
            "Provide a non-negative overlap duration value."
        )
    
    if overlap_duration >= window_duration:
        raise BreesyError(
            f"Overlap duration ({overlap_duration}s) must be less than window duration ({window_duration}s).",
            "Provide an overlap duration that is smaller than the window duration."
        )

    num_samples_per_epoch = int(window_duration * recording.sample_rate)
    
    if num_samples_per_epoch > recording.number_of_samples:
        raise BreesyError(
            f"Epoch duration ({window_duration}s) is longer than the recording duration ({recording.duration}s).",
            "Make sure that the epoch duration is shorter than the recording duration. "
            "You can see the recording duration like so: recording.duration."
        )

    step_duration = window_duration - overlap_duration
    step_samples = int(step_duration * recording.sample_rate)
    
    max_start_sample = recording.number_of_samples - num_samples_per_epoch
    num_epochs = (max_start_sample // step_samples) + 1

    epoch_recordings = []
    for i in range(num_epochs):
        start_index = i * step_samples
        end_index = start_index + num_samples_per_epoch
        
        # skip window if out of bounds
        if end_index > recording.number_of_samples:
            break
            
        epoch_data = recording.data[:, start_index:end_index]
        epoch_recording = Recording(
            data=epoch_data,
            channel_names=recording.channel_names,
            sample_rate=recording.sample_rate,
            events=recording.events,
            metadata=recording.metadata
        )
        epoch_recordings.append(epoch_recording)

    return epoch_recordings


# TODO: update function to FIR and unhardcode
@enforce_type_hints
def select_bandwidth(recording: Recording, low: float | int, high: float | int, filter_order: int = 5) -> Recording:
    """Filter recording to keep only a specific frequency band.

    :param recording: Input recording
    :param low: Lower frequency bound in Hz
    :param high: Upper frequency bound in Hz
    :param filter_order: Order of the filter
    
    :return: Filtered recording
    """
    data = recording.data
    sample_rate = recording.sample_rate

    low_high_range_valid = low <= high
    if not low_high_range_valid:
        raise BreesyError("Low frequency bound is higher than the high frequency bound",
                          "Provide a low frequency bound value that is less than or equal to the high frequency bound")

    high_bound_exceeds_nyquist = high >= (sample_rate / 2)
    if high_bound_exceeds_nyquist:
        raise BreesyError("High frequency bound exceeds Nyquist frequency",
                          "Provide a high frequency bound value that is less than half of the sample rate")

    bandpass = _si_butter_bandpass_sos(low, high, sample_rate, filter_order)

    new_eeg_data = _si_sosfiltfilt(bandpass, data)

    return update_data(recording, new_eeg_data)


from urllib.request import urlopen, Request
import certifi, ssl
from shutil import copyfileobj
from pathlib import Path
from scipy.io import loadmat
# from mne.io import read_raw_edf

from .Recording import Recording
from .RecordingMetadata import RecordingMetadata
from .errors import BreesyError


def load_example_data(dataset_name: str, dir_name: str = "tmp", ssl_context=None) -> Recording:
    """Load example EEG datasets for testing and learning.

    Downloads and loads publicly available EEG data files. Currently available datasets include:
    'alpha', 'madhd', 'pangolin'.

    :param dataset_name: Name of the dataset to load
    :param dir_name: Directory to save downloaded files
    :param ssl_context: SSL context for secure downloads

    :return: Recording object containing the loaded example data
    """

    available_datasets = ['alpha', 'madhd', 'pangolin']

    # TODO: some downloads take time, do some loading bar?
    if dataset_name == "alpha":
        filename = "subject_01.mat"
        url = "https://zenodo.org/records/2348892/files/subject_01.mat?download=1"
        ch_names = ["Fp1", "Fp2", "FC5", "FC6", "Fz", "T7", "Cz", "T8",
                    "P7", "P3", "Pz", "P4", "P8", "O1", "Oz", "O2"]
        sample_rate = 512

    elif dataset_name == "madhd":
        filename = "MADHD.mat"
        url = "https://data.mendeley.com/public-files/datasets/6k4g25fhzg/files/9c5928cf-f8ef-485b-ac5b-f17b2df935f3/file_downloaded"
        ch_names = ["Cz", "F4"]
        sample_rate = 256

    # elif dataset_name == "schalk":
    #     filename = "S001R01.edf"
    #     url = "https://www.physionet.org/files/eegmmidb/1.0.0/S001/S001R01.edf?download"
    #     ch_names = ['Fc5', 'Fc3', 'Fc1', 'Fcz', 'Fc2', 'Fc4', 'Fc6',
    #                 'C5', 'C3', 'C1', 'Cz', 'C2', 'C4', 'C6',
    #                 'Cp5', 'Cp3', 'Cp1', 'Cpz', 'Cp2', 'Cp4', 'Cp6',
    #                 'Fp1', 'Fpz', 'Fp2', 'Af7', 'Af3', 'Afz', 'Af4', 'Af8',
    #                 'F7', 'F5', 'F3', 'F1', 'Fz', 'F2', 'F4', 'F6', 'F8',
    #                 'Ft7', 'Ft8', 'T7', 'T8', 'T9', 'T10', 'Tp7', 'Tp8',
    #                 'P7', 'P5', 'P3', 'P1', 'Pz', 'P2', 'P4', 'P6', 'P8',
    #                 'Po7', 'Po3', 'Poz', 'Po4', 'Po8', 'O1', 'Oz', 'O2', 'Iz'
    #                 ]
    #     sample_rate = 160

    elif dataset_name == "pangolin":
        filename = "S1_run8.mat"
        url = "https://osf.io/download/jrcn8/?view_only=d23acfd50655427fbaae381a17cbfbcc"
        ch_names = [f"ch{i+1}" for i in range(258)]  # this dataset has no standard channel names
        sample_rate = 600

    # elif dataset_name == "telemetry":
    #     filename = "ST7011J0-PSG.edf"
    #     url = "https://www.physionet.org/files/sleep-edfx/1.0.0/sleep-telemetry/ST7011J0-PSG.edf?download"
    #     ch_names = range(5)  # TODO: get actual channel names
    #     sample_rate = 100

    else:
        available_datasets_string = ", ".join([f'"{n}"' for n in available_datasets])
        raise BreesyError(f'Unknown name of dataset provided: "{dataset_name}".', 
                          f'Use one of available dataset names ({available_datasets_string}).')

    dirpath = Path(dir_name)
    dirpath.mkdir(parents=True, exist_ok=True)
    filepath = dirpath / filename

    if not filepath.exists():  # TODO: use temp download location instead of working dir
        print(f'Will download the data file from {url}')
        if ssl_context is None:
            ssl_context = ssl.create_default_context(cafile=certifi.where())
        req = Request(url, headers={'User-Agent' : "Magic Browser"}) 
        with urlopen(req, context=ssl_context) as in_stream, open(filepath, 'wb') as out_file:
            copyfileobj(in_stream, out_file)
        print(f'Done! Save file in: {filepath}')

    if dataset_name == "alpha":
        data = loadmat(filepath)["SIGNAL"].T[1:-2]  # removes non-EEG channels
    elif dataset_name == "madhd":
        data = loadmat(filepath, squeeze_me=True)['MADHD'][0][0, :, :].T  # leaves one recording only
    elif dataset_name == "pangolin":
        data = loadmat(filepath, squeeze_me=True)['y']
    elif dataset_name in ["schalk", "telemetry"]:
        # data = read_raw_edf(filepath, verbose=0).get_data()  # TODO: remove MNE requirement
        raise ValueError(f'The "{dataset_name}" is temporarily unsupported as it is an EDF file and requires MNE to be loaded.')

    metadata = RecordingMetadata(
        name=Path(filename).stem,
        file_path=filepath,
        description=f"Sample dataset: {dataset_name}"
    )

    recording = Recording(
        data=data,
        channel_names=ch_names,
        sample_rate=sample_rate,
        events=None,
        metadata=metadata
    )

    return recording